@@ -1,28 +1,42 @@
+from __future__ import annotations
+
 import asyncio
 import functools
 import logging
 import socket
 import sys
-from collections.abc import Iterable, Sequence
+import warnings
+import weakref
+from collections.abc import Callable, Iterable, Sequence
 from types import TracebackType
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Callable,
-    Literal,
-    Optional,
-    TypeVar,
-    Union,
-    overload,
-)
+from typing import TYPE_CHECKING, Any, Literal, TypeVar, overload
 
 import pycares
 
 from . import error
+from .compat import (
+    AresHostResult,
+    AresQueryAAAAResult,
+    AresQueryAResult,
+    AresQueryCAAResult,
+    AresQueryCNAMEResult,
+    AresQueryMXResult,
+    AresQueryNAPTRResult,
+    AresQueryNSResult,
+    AresQueryPTRResult,
+    AresQuerySOAResult,
+    AresQuerySRVResult,
+    AresQueryTXTResult,
+    QueryResult,
+    convert_result,
+)
 
-__version__ = '3.6.1'
+__version__ = '4.0.0'
 
-__all__ = ('DNSResolver', 'error')
+__all__ = (
+    'DNSResolver',
+    'error',
+)
 
 _T = TypeVar('_T')
 
@@ -60,8 +74,8 @@ query_class_map = {
 class DNSResolver:
     def __init__(
         self,
-        nameservers: Optional[Sequence[str]] = None,
-        loop: Optional[asyncio.AbstractEventLoop] = None,
+        nameservers: Sequence[str] | None = None,
+        loop: asyncio.AbstractEventLoop | None = None,
         **kwargs: Any,
     ) -> None:  # TODO(PY311): Use Unpack for kwargs.
         self._closed = True
@@ -76,36 +90,32 @@ class DNSResolver:
             self.nameservers = nameservers
         self._read_fds: set[int] = set()
         self._write_fds: set[int] = set()
-        self._timer: Optional[asyncio.TimerHandle] = None
+        self._timer: asyncio.TimerHandle | None = None
         self._closed = False
 
     def _make_channel(self, **kwargs: Any) -> tuple[bool, pycares.Channel]:
-        if (
-            hasattr(pycares, 'ares_threadsafety')
-            and pycares.ares_threadsafety()
-        ):
-            # pycares is thread safe
-            try:
-                return True, pycares.Channel(
-                    event_thread=True, timeout=self._timeout, **kwargs
+        # pycares 5+ uses event_thread by default when sock_state_cb
+        # is not provided
+        try:
+            return True, pycares.Channel(timeout=self._timeout, **kwargs)
+        except pycares.AresError as e:
+            if sys.platform == 'linux':
+                _LOGGER.warning(
+                    'Failed to create DNS resolver channel with automatic '
+                    'monitoring of resolver configuration changes. This '
+                    'usually means the system ran out of inotify watches. '
+                    'Falling back to socket state callback. Consider '
+                    'increasing the system inotify watch limit: %s',
+                    e,
                 )
-            except pycares.AresError as e:
-                if sys.platform == 'linux':
-                    _LOGGER.warning(
-                        'Failed to create DNS resolver channel with automatic '
-                        'monitoring of resolver configuration changes. This '
-                        'usually means the system ran out of inotify watches. '
-                        'Falling back to socket state callback. Consider '
-                        'increasing the system inotify watch limit: %s',
-                        e,
-                    )
-                else:
-                    _LOGGER.warning(
-                        'Failed to create DNS resolver channel with automatic '
-                        'monitoring of resolver configuration changes. '
-                        'Falling back to socket state callback: %s',
-                        e,
-                    )
+            else:
+                _LOGGER.warning(
+                    'Failed to create DNS resolver channel with automatic '
+                    'monitoring of resolver configuration changes. '
+                    'Falling back to socket state callback: %s',
+                    e,
+                )
+        # Fall back to sock_state_cb (needs SelectorEventLoop on Windows)
         if sys.platform == 'win32' and not isinstance(
             self.loop, asyncio.SelectorEventLoop
         ):
@@ -116,20 +126,37 @@ class DNSResolver:
                     raise RuntimeError(WINDOWS_SELECTOR_ERR_MSG)
             except ModuleNotFoundError as ex:
                 raise RuntimeError(WINDOWS_SELECTOR_ERR_MSG) from ex
+        # Use weak reference for deterministic cleanup. Without it there's a
+        # reference cycle (DNSResolver -> _channel -> callback -> DNSResolver).
+        # Python 3.4+ can handle cycles with __del__, but weak ref ensures
+        # cleanup happens immediately when last reference is dropped.
+        weak_self = weakref.ref(self)
+
+        def sock_state_cb_wrapper(
+            fd: int, readable: bool, writable: bool
+        ) -> None:
+            this = weak_self()
+            if this is not None:
+                this._sock_state_cb(fd, readable, writable)
+
         return False, pycares.Channel(
-            sock_state_cb=self._sock_state_cb, timeout=self._timeout, **kwargs
+            sock_state_cb=sock_state_cb_wrapper,
+            timeout=self._timeout,
+            **kwargs,
         )
 
     @property
     def nameservers(self) -> Sequence[str]:
-        return self._channel.servers
+        # pycares 5.x returns servers with port (e.g., '8.8.8.8:53')
+        # Strip port for backward compatibility with pycares 4.x
+        return [s.rsplit(':', 1)[0] for s in self._channel.servers]
 
     @nameservers.setter
-    def nameservers(self, value: Iterable[Union[str, bytes]]) -> None:
+    def nameservers(self, value: Iterable[str | bytes]) -> None:
         self._channel.servers = value
 
     def _callback(
-        self, fut: asyncio.Future[_T], result: _T, errorno: Optional[int]
+        self, fut: asyncio.Future[_T], result: _T, errorno: int | None
     ) -> None:
         if fut.cancelled():
             return
@@ -142,9 +169,9 @@ class DNSResolver:
 
     def _get_future_callback(
         self,
-    ) -> tuple['asyncio.Future[_T]', Callable[[_T, int], None]]:
+    ) -> tuple[asyncio.Future[_T], Callable[[_T, int | None], None]]:
         """Return a future and a callback to set the result of the future."""
-        cb: Callable[[_T, int], None]
+        cb: Callable[[_T, int | None], None]
         future: asyncio.Future[_T] = self.loop.create_future()
         if self._event_thread:
             cb = functools.partial(  # type: ignore[assignment]
@@ -156,109 +183,227 @@ class DNSResolver:
             cb = functools.partial(self._callback, future)
         return future, cb
 
+    def _query_callback(
+        self,
+        fut: asyncio.Future[QueryResult],
+        qtype: int,
+        result: pycares.DNSResult,
+        errorno: int | None,
+    ) -> None:
+        """Callback for query that converts results to compatible format."""
+        if fut.cancelled():
+            return
+        if errorno is not None:
+            fut.set_exception(
+                error.DNSError(errorno, pycares.errno.strerror(errorno))
+            )
+        else:
+            fut.set_result(convert_result(result, qtype))
+
+    def _get_query_future_callback(
+        self, qtype: int
+    ) -> tuple[asyncio.Future[QueryResult], Callable[..., None]]:
+        """Return a future and callback for query with result conversion."""
+        future: asyncio.Future[QueryResult] = self.loop.create_future()
+        cb: Callable[..., None]
+        if self._event_thread:
+            cb = functools.partial(  # type: ignore[assignment]
+                self.loop.call_soon_threadsafe,
+                self._query_callback,  # type: ignore[arg-type]
+                future,
+                qtype,
+            )
+        else:
+            cb = functools.partial(self._query_callback, future, qtype)
+        return future, cb
+
     @overload
     def query(
-        self, host: str, qtype: Literal['A'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[list[pycares.ares_query_a_result]]: ...
+        self, host: str, qtype: Literal['A'], qclass: str | None = ...
+    ) -> asyncio.Future[list[AresQueryAResult]]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['AAAA'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[list[pycares.ares_query_aaaa_result]]: ...
+        self, host: str, qtype: Literal['AAAA'], qclass: str | None = ...
+    ) -> asyncio.Future[list[AresQueryAAAAResult]]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['CAA'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[list[pycares.ares_query_caa_result]]: ...
+        self, host: str, qtype: Literal['CAA'], qclass: str | None = ...
+    ) -> asyncio.Future[list[AresQueryCAAResult]]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['CNAME'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[pycares.ares_query_cname_result]: ...
+        self, host: str, qtype: Literal['CNAME'], qclass: str | None = ...
+    ) -> asyncio.Future[AresQueryCNAMEResult]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['MX'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[list[pycares.ares_query_mx_result]]: ...
+        self, host: str, qtype: Literal['MX'], qclass: str | None = ...
+    ) -> asyncio.Future[list[AresQueryMXResult]]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['NAPTR'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[list[pycares.ares_query_naptr_result]]: ...
+        self, host: str, qtype: Literal['NAPTR'], qclass: str | None = ...
+    ) -> asyncio.Future[list[AresQueryNAPTRResult]]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['NS'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[list[pycares.ares_query_ns_result]]: ...
+        self, host: str, qtype: Literal['NS'], qclass: str | None = ...
+    ) -> asyncio.Future[list[AresQueryNSResult]]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['PTR'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[list[pycares.ares_query_ptr_result]]: ...
+        self, host: str, qtype: Literal['PTR'], qclass: str | None = ...
+    ) -> asyncio.Future[AresQueryPTRResult]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['SOA'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[pycares.ares_query_soa_result]: ...
+        self, host: str, qtype: Literal['SOA'], qclass: str | None = ...
+    ) -> asyncio.Future[AresQuerySOAResult]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['SRV'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[list[pycares.ares_query_srv_result]]: ...
+        self, host: str, qtype: Literal['SRV'], qclass: str | None = ...
+    ) -> asyncio.Future[list[AresQuerySRVResult]]: ...
     @overload
     def query(
-        self, host: str, qtype: Literal['TXT'], qclass: Optional[str] = ...
-    ) -> asyncio.Future[list[pycares.ares_query_txt_result]]: ...
+        self, host: str, qtype: Literal['TXT'], qclass: str | None = ...
+    ) -> asyncio.Future[list[AresQueryTXTResult]]: ...
 
     def query(
-        self, host: str, qtype: str, qclass: Optional[str] = None
-    ) -> Union[asyncio.Future[list[Any]], asyncio.Future[Any]]:
+        self, host: str, qtype: str, qclass: str | None = None
+    ) -> asyncio.Future[list[Any]] | asyncio.Future[Any]:
+        """Query DNS records (deprecated, use query_dns instead)."""
+        warnings.warn(
+            'query() is deprecated, use query_dns() instead',
+            DeprecationWarning,
+            stacklevel=2,
+        )
         try:
-            qtype = query_type_map[qtype]
+            qtype_int = query_type_map[qtype]
         except KeyError as e:
             raise ValueError(f'invalid query type: {qtype}') from e
+        qclass_int: int | None = None
         if qclass is not None:
             try:
-                qclass = query_class_map[qclass]
+                qclass_int = query_class_map[qclass]
             except KeyError as e:
                 raise ValueError(f'invalid query class: {qclass}') from e
 
-        fut: Union[asyncio.Future[list[Any]], asyncio.Future[Any]]
-        fut, cb = self._get_future_callback()
-        self._channel.query(host, qtype, cb, query_class=qclass)
+        fut, cb = self._get_query_future_callback(qtype_int)
+        if qclass_int is not None:
+            self._channel.query(
+                host, qtype_int, query_class=qclass_int, callback=cb
+            )
+        else:
+            self._channel.query(host, qtype_int, callback=cb)
         return fut
 
+    def query_dns(
+        self, host: str, qtype: str, qclass: str | None = None
+    ) -> asyncio.Future[pycares.DNSResult]:
+        """Query DNS records, returning native pycares 5.x DNSResult."""
+        try:
+            qtype_int = query_type_map[qtype]
+        except KeyError as e:
+            raise ValueError(f'invalid query type: {qtype}') from e
+        qclass_int: int | None = None
+        if qclass is not None:
+            try:
+                qclass_int = query_class_map[qclass]
+            except KeyError as e:
+                raise ValueError(f'invalid query class: {qclass}') from e
+
+        fut: asyncio.Future[pycares.DNSResult]
+        fut, cb = self._get_future_callback()
+        if qclass_int is not None:
+            self._channel.query(
+                host, qtype_int, query_class=qclass_int, callback=cb
+            )
+        else:
+            self._channel.query(host, qtype_int, callback=cb)
+        return fut
+
+    def _gethostbyname_callback(
+        self,
+        fut: asyncio.Future[AresHostResult],
+        host: str,
+        result: pycares.AddrInfoResult | None,
+        errorno: int | None,
+    ) -> None:
+        """Callback for gethostbyname that converts AddrInfoResult."""
+        if fut.cancelled():
+            return
+        if errorno is not None:
+            fut.set_exception(
+                error.DNSError(errorno, pycares.errno.strerror(errorno))
+            )
+        else:
+            assert result is not None  # noqa: S101
+            # node.addr is (address_bytes, port) - extract and decode
+            addresses = [node.addr[0].decode() for node in result.nodes]
+            # Get canonical name from cnames if available
+            name = result.cnames[0].name if result.cnames else host
+            fut.set_result(
+                AresHostResult(name=name, aliases=[], addresses=addresses)
+            )
+
     def gethostbyname(
         self, host: str, family: socket.AddressFamily
-    ) -> asyncio.Future[pycares.ares_host_result]:
-        fut: asyncio.Future[pycares.ares_host_result]
-        fut, cb = self._get_future_callback()
-        self._channel.gethostbyname(host, family, cb)
+    ) -> asyncio.Future[AresHostResult]:
+        """
+        Resolve hostname to addresses.
+
+        Deprecated: Use getaddrinfo() instead. This is implemented using
+        getaddrinfo as pycares 5.x removed the gethostbyname method.
+        """
+        warnings.warn(
+            'gethostbyname() is deprecated, use getaddrinfo() instead',
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        fut: asyncio.Future[AresHostResult] = self.loop.create_future()
+        cb: Callable[..., None]
+        if self._event_thread:
+            cb = functools.partial(  # type: ignore[assignment]
+                self.loop.call_soon_threadsafe,
+                self._gethostbyname_callback,  # type: ignore[arg-type]
+                fut,
+                host,
+            )
+        else:
+            cb = functools.partial(self._gethostbyname_callback, fut, host)
+        self._channel.getaddrinfo(host, None, family=family, callback=cb)
         return fut
 
     def getaddrinfo(
         self,
         host: str,
         family: socket.AddressFamily = socket.AF_UNSPEC,
-        port: Optional[int] = None,
+        port: int | None = None,
         proto: int = 0,
         type: int = 0,
         flags: int = 0,
-    ) -> asyncio.Future[pycares.ares_addrinfo_result]:
-        fut: asyncio.Future[pycares.ares_addrinfo_result]
+    ) -> asyncio.Future[pycares.AddrInfoResult]:
+        fut: asyncio.Future[pycares.AddrInfoResult]
         fut, cb = self._get_future_callback()
         self._channel.getaddrinfo(
-            host, port, cb, family=family, type=type, proto=proto, flags=flags
+            host,
+            port,
+            family=family,
+            type=type,
+            proto=proto,
+            flags=flags,
+            callback=cb,
         )
         return fut
 
     def getnameinfo(
         self,
-        sockaddr: Union[tuple[str, int], tuple[str, int, int, int]],
+        sockaddr: tuple[str, int] | tuple[str, int, int, int],
         flags: int = 0,
-    ) -> asyncio.Future[pycares.ares_nameinfo_result]:
-        fut: asyncio.Future[pycares.ares_nameinfo_result]
+    ) -> asyncio.Future[pycares.NameInfoResult]:
+        fut: asyncio.Future[pycares.NameInfoResult]
         fut, cb = self._get_future_callback()
-        self._channel.getnameinfo(sockaddr, flags, cb)
+        self._channel.getnameinfo(sockaddr, flags, callback=cb)
         return fut
 
-    def gethostbyaddr(
-        self, name: str
-    ) -> asyncio.Future[pycares.ares_host_result]:
-        fut: asyncio.Future[pycares.ares_host_result]
+    def gethostbyaddr(self, name: str) -> asyncio.Future[pycares.HostResult]:
+        fut: asyncio.Future[pycares.HostResult]
         fut, cb = self._get_future_callback()
-        self._channel.gethostbyaddr(name, cb)
+        self._channel.gethostbyaddr(name, callback=cb)
         return fut
 
     def cancel(self) -> None:
@@ -342,17 +487,19 @@ class DNSResolver:
         This should be called to ensure all resources are properly released.
         After calling close(), the resolver should not be used again.
         """
+        if not self._closed:
+            self._channel.cancel()
         self._cleanup()
 
-    async def __aenter__(self) -> 'DNSResolver':
+    async def __aenter__(self) -> DNSResolver:
         """Enter the async context manager."""
         return self
 
     async def __aexit__(
         self,
-        exc_type: Optional[type[BaseException]],
-        exc_val: Optional[BaseException],
-        exc_tb: Optional[TracebackType],
+        exc_type: type[BaseException] | None,
+        exc_val: BaseException | None,
+        exc_tb: TracebackType | None,
     ) -> None:
         """Exit the async context manager."""
         await self.close()
